﻿open System.Threading

(* Raima.Rdm is the system module that contains the ADO.NET data provider
   for RDM *)
open Raima.Rdm

(* EXAMPLE - SQL11 for F# using ADO.NET *)
(* This example shows multi-user access to a database using separate threads.
 * Two writer threads are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader thread is also created
 * that prints out the current oldest and newest entries every second.  Each
 * thread uses regular locking.
 * 
 * The following methods are illustrated:
 *     Raima.Rdm.RdmConnection:
 *         ChangeDatabase(string) Method
 *         Close() Method
 *         CreateCommand() Method
 *         Open() Method
 *         RdmConnection(String) Constructor
 *     Raima.Rdm.RdmCommand:
 *         CommandText Property
 *         ExecuteNonQuery() Method
 *         ExecuteReader() Method
 *         Parameters Property
 *     Raima.Rdm.RdmDataReader:
 *         Item[int] Property
 *         Read() Method
 *     Raima.Rdm.RdmParameterCollection
 *         Add(string, RdmType)
 *         Item[string] Property
 * 
 * They are fully document in http://docs.raima.com/rdm/.
 * 
 * For simplicity this example does not catch any thrown exceptions
 * but good programming practices would dictate that this occur in
 * a full application.
 *)

let mutable shutdown = false

let DropDatabase (conn : RdmConnection) name =
    let cmd = conn.CreateCommand()

    (* Attempt to drop this database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. *)
    try
        cmd.CommandText <- "DROP DATABASE " + name
        cmd.ExecuteNonQuery() |> ignore
    with
        | exn -> ()

let CreateThisDatabase (conn : RdmConnection) name cmds =
    let cmd = conn.CreateCommand()

    cmd.CommandText <- "CREATE DATABASE " + name
    cmd.ExecuteNonQuery() |> ignore

    cmds |>
        Array.iter(fun c ->
            cmd.CommandText <- c
            cmd.ExecuteNonQuery() |> ignore
        )

    cmd.CommandText <- "COMMIT"
    cmd.ExecuteNonQuery() |> ignore

    cmd.CommandText <- "CLOSE " + name
    cmd.ExecuteNonQuery() |> ignore

let CreateDatabases conn =
    (* Drop any existing copies of the databases. *)
    DropDatabase conn "SQL11"

    (* Attempt to create the databases *)
    CreateThisDatabase conn "SQL11" [| "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)" |]

let InserterThreadProc _ =
    (* Create a connection for this thread *)
    let conn = new RdmConnection("host=local;database=SQL11;openmode=s")
    conn.Open()

    (* Create a command to do the insert *)
    let cmd = conn.CreateCommand()
    cmd.CommandText <- "INSERT INTO Table1 VALUES :id"
    cmd.Parameters.Add("id", RdmType.Int32) |> ignore

    (* Loop until a shutdown is requested *)
    let seq = Seq.unfold(fun ii -> if shutdown then None else Some(ii, ii + 1)) 0
    
    seq |>
        Seq.iter(fun ii ->
            (* Insert the next value *)
            cmd.Parameters.["id"].Value <- ii
            cmd.ExecuteNonQuery() |> ignore)

    conn.Close()
    ()
    
let DeleterThreadProc _ =
    (* Create a connection for this thread *)
    let conn = new RdmConnection("host=local;database=SQL11;openmode=s")
    conn.Open()

    (* Create a command to check the number of rows, one to find the
     * 'smallest' row, and one to delete *)
    let chkCmd = conn.CreateCommand()
    let minCmd = conn.CreateCommand()
    let delCmd = conn.CreateCommand()

    chkCmd.CommandText <- "SELECT COUNT(id) FROM Table1"
    minCmd.CommandText <- "SELECT MIN(id) FROM Table1"
    delCmd.CommandText <- "DELETE FROM Table1 WHERE id = :minid"
    let parm = delCmd.Parameters.Add("minid", RdmType.Int32)

    (* Loop until a shutdown is requested *)
    while shutdown = false do
        (* Query the number of rows in the table *)
        let countRdr = chkCmd.ExecuteReader()
        if countRdr.Read() then
            let count = System.Convert.ToInt64(countRdr.[0])

            (* There should be only one row, so this closes the reader *)
            countRdr.Read() |> ignore

            if (count > 100L) then
                (* If there are at least 100 rows, get the smallest and delete it *)
                let minRdr = minCmd.ExecuteReader()
                if minRdr.Read() then
                    parm.Value <- System.Convert.ToInt32(minRdr.[0])
                    delCmd.ExecuteNonQuery() |> ignore

                    (* There should be only one row, so this closes the reader *)
                    minRdr.Read() |> ignore

    conn.Close()

let ReaderThreadProc _ =
    (* Create a connection for this thread *)
    let conn = new RdmConnection("host=local;database=SQL11;openmode=s")
    conn.Open()

    (* Create a command to do the read *)
    let cmd = conn.CreateCommand()
    cmd.CommandText <- "SELECT MIN(id), MAX(id) FROM Table1"

    (* Loop until a shutdown is requested *)
    while shutdown = false do
        (* Wait one second *)
        Thread.Sleep(1000)

        (* Query the minimum and maximum values *)
        let rdr = cmd.ExecuteReader()
        if rdr.Read() then
            printfn "Current minimum = %A and current minimum = %A" rdr.[0] rdr.[1]
            rdr.Read() |> ignore

    conn.Close()

(* Create the connection with a simple connection string *)
let conn = new RdmConnection("host=local;openmode=s")
conn.Open()

CreateDatabases conn

(* Create three threads, one to insert, one to delete, and one
 * to show the current contents *)
let inserter = new Thread(new ThreadStart(InserterThreadProc))
let deleter = new Thread(new ThreadStart(DeleterThreadProc))
let reader = new Thread(new ThreadStart(ReaderThreadProc))

printfn "Starting threads"
printfn "   Press <Enter> when you are ready to terminate them"

(* Start each of the threads *)
inserter.Start()
deleter.Start()
reader.Start()

(* Wait for the user to hit <Enter> *)
System.Console.ReadLine() |> ignore

(*Indicate to the threads that they should shut down *)
shutdown <- true

(*Wait for each thread to finish *)
inserter.Join()
deleter.Join()
reader.Join()

printfn "All threads have terminated"

conn.Close()

printfn "Hit <Enter> to continue"
System.Console.ReadLine() |> ignore

